@@ -4,17 +4,27 @@ Created on Tue Jan 25 16:34:41 2022
 @authors: Andrea Bassi @Polimi, Mark Neil @ImperialCollege
 """
 from napari_sim_processor.widget_settings import Setting
+from napari_sim_processor.baseSimProcessor import pytorch, cupy
 from napari_sim_processor.hexSimProcessor import HexSimProcessor
 from napari_sim_processor.convSimProcessor import ConvSimProcessor
-from napari_sim_processor.simProcessor import SimProcessor 
+from napari_sim_processor.simProcessor import SimProcessor
 import napari
-from qtpy.QtWidgets import QVBoxLayout,QSplitter, QHBoxLayout, QWidget, QPushButton
+from qtpy.QtWidgets import QVBoxLayout,QSplitter, QHBoxLayout, QWidget, QPushButton, QComboBox, QLineEdit
 from napari.layers import Image
 import numpy as np
 from napari.qt.threading import thread_worker
 from magicgui.widgets import FunctionGui
 from magicgui import magicgui, magic_factory
 import warnings
+import enum
+import time
+from superqt.utils import qthrottled
+
+
+class accel(enum.Enum):
+    USENUMPY = 1
+    USETORCH = 2
+    USECUPY = 3
 
 
 def reshape_init(reshape_widget: FunctionGui):
@@ -40,7 +50,7 @@ def reshape_init(reshape_widget: FunctionGui):
                call_button="Reshape stack",
                x={'max': 4096},
                y={'max': 4096},
-               input_order={"choices": ['apzyx', 'pazyx', 'pzayx', 'azpyx']},
+               input_order={"choices": ['apzyx', 'pazyx', 'pzayx', 'azpyx', 'zapyx', 'zpayx']},
                )
 def reshape(viewer: napari.Viewer,
             input_image: Image,
@@ -84,11 +94,16 @@ def reshape(viewer: napari.Viewer,
         elif input_order == 'pzayx':
             rdata = np.moveaxis(data.reshape(phases, z, angles, y, x),[0,1,2],[1,2,0])
         elif input_order == 'azpyx':
-            rdata = np.moveaxis(data.reshape(angles, z, phases, y, x),[1,2],[2,1])
+            rdata = np.moveaxis(data.reshape(angles, z, phases, y, x), [1, 2], [2, 1])
+        elif input_order == 'zapyx':
+            rdata = np.moveaxis(data.reshape(z, angles, phases, y, x), [0, 1, 2], [2, 0, 1])
+        elif input_order == 'zpayx':
+            rdata = np.moveaxis(data.reshape(z, phases, angles, y, x), [0, 2], [2, 0])
         else:
             raise(ValueError('Input stack order reshaping not implemented'))
         input_image.data = rdata
         viewer.dims.axis_labels = ["angle", "phase", "z", "y","x"]
+        viewer.dims.set_point(axis=[0,1,2], value=[0,0,0]) #TODO  check why this is raising ValueError sistematically 
 
 
 class SimAnalysis(QWidget):
@@ -96,7 +111,7 @@ class SimAnalysis(QWidget):
     A Napari plugin for the reconstruction of Structured Illumination microscopy (SIM) data with GPU acceleration (with pytorch, if installed).
     Currently supports:    
    - conventional data with improved resolution in 1D (1 angle, 3 phases)
-   - conventional data (3 angles, 3 phases)
+   - conventional data with a generic number of angles and phases
    - hexagonal SIM (1 angle, 7 phases).
     Accepts image stacks organized in 5D (angle,phase,z,y,x).
     For stacks with multiple z-frames each plane is processed as described in:
@@ -158,13 +173,13 @@ class SimAnalysis(QWidget):
         self.beta = Setting('beta', dtype=float, initial=0.980, spinbox_step=0.01, 
                              layout=left_layout,  spinbox_decimals=3,
                              write_function = self.setReconstructor)
-        self.w = Setting('w', dtype=float, initial=0.5, layout=left_layout,
+        self.w = Setting('w', dtype=float, initial=0.2, layout=left_layout,
                               spinbox_decimals=2,
                               write_function = self.setReconstructor)
         self.eta = Setting('eta', dtype=float, initial=0.65,
                             layout=left_layout, spinbox_decimals=3, spinbox_step=0.01,
                             write_function = self.setReconstructor)
-        self.group = Setting('group', dtype=int, initial=30, vmin=2,
+        self.group = Setting('group', dtype=int, initial=30, vmin=1,
                             layout=left_layout,
                             write_function = self.setReconstructor)
         self.use_phases = Setting('use phases', dtype=bool, initial=True, layout=left_layout,                         
@@ -173,6 +188,10 @@ class SimAnalysis(QWidget):
                                      layout=left_layout, 
                                      write_function = self.setReconstructor) 
         # Fill bottom-right layout    
+        self.carrier_idx = Setting('carrier index', dtype=int, initial=0,
+                                   layout=right_layout, vmin = 0,
+                                   write_function = self.show_xcorr
+                                   )
         self.showXcorr = Setting('Show Xcorr', dtype=bool, initial=False,
                                      layout=right_layout,
                                      write_function = self.show_xcorr
@@ -196,14 +215,18 @@ class SimAnalysis(QWidget):
         self.keep_calibrating = Setting('Continuos Calibration', dtype=bool, initial=False,
                                      layout=right_layout, 
                                      write_function = self.setReconstructor)
-        # self.keep_reconstructing = Setting('Continuos Reconstruction', dtype=bool, initial=False,
-        #                              layout=right_layout, 
-        #                              write_function = self.setReconstructor)
         self.batch = Setting('Batch Reconstruction', dtype=bool, initial=False,
                                      layout=right_layout, 
                                      write_function = self.setReconstructor)
-        self.use_torch = Setting('Use Torch', dtype=bool, initial=False, layout=right_layout, 
-                           write_function = self.setReconstructor)         
+
+        self.proc = QComboBox()
+        self.proc.addItem('Use Numpy', userData=accel.USENUMPY)
+        if pytorch:
+            self.proc.addItem('Use PyTorch', userData=accel.USETORCH)
+        if cupy:
+            self.proc.addItem('Use CuPy', userData=accel.USECUPY)
+        right_layout.addWidget(self.proc)
+
         buttons_dict = {'Widefield': self.calculate_WF_image,
                         'Calibrate': self.calibration,
                         'Plot calibration phases':self.find_phaseshifts,
@@ -214,8 +237,11 @@ class SimAnalysis(QWidget):
         for button_name, call_function in buttons_dict.items():
             button = QPushButton(button_name)
             button.clicked.connect(call_function)
-            right_layout.addWidget(button) 
-    
+            right_layout.addWidget(button)
+
+        self.messageBox = QLineEdit()
+        layout.addWidget(self.messageBox, stretch=True)
+        self.messageBox.setText('Messages')
         
     def add_magic_layer_selection(self,_layout):
         function = self.select_layer
@@ -285,14 +311,21 @@ class SimAnalysis(QWidget):
                 current_step[dim_idx] = data.shape[dim_idx]//2
             self.viewer.dims.current_step = current_step                
            
-            
+    @qthrottled (timeout=10)
     def on_step_change(self, *args):   
-        if hasattr(self, 'imageRaw_name'): 
+        if hasattr(self, 'imageRaw_name'):
+            t0 = time.time()
             self.setReconstructor()
             self.show_spectrum()
             self.show_xcorr()
-            
-                 
+            min_timeout = 10 #ms
+            delta_t = time.time() - t0
+            self.set_timeout(delta_t + min_timeout)
+
+    def set_timeout(self, new_timeout):
+        self.on_step_change.set_timeout(new_timeout)
+
+
     def show_image(self, image_values, im_name, **kwargs):
         '''
         creates a new Image layer with image_values as data
@@ -369,7 +402,7 @@ class SimAnalysis(QWidget):
         try:
             return self.viewer.layers[self.imageRaw_name].data
         except:
-             raise(KeyError('Please select a valid stack'))
+            raise(KeyError('Please select a valid stack'))
     
     
     def get_current_ap_stack(self):
@@ -429,9 +462,10 @@ class SimAnalysis(QWidget):
             elif self.phases_number.val == 7:  
                 self.h = HexSimProcessor()  
                 k_shape = (3,1)
-            elif self.phases_number.val == 3 and self.angles_number.val == 3: 
-                self.h = ConvSimProcessor()
-                k_shape = (3,1)
+            elif self.phases_number.val >= 3 and self.angles_number.val > 0:
+                self.h = ConvSimProcessor(angleSteps=self.angles_number.val,
+                                          phaseSteps=self.phases_number.val)
+                k_shape = (self.angles_number.val,1)
             else: 
                 raise(ValueError("Invalid phases or angles number"))
             self.h.debug = False
@@ -524,14 +558,22 @@ class SimAnalysis(QWidget):
         if self.is_image_in_layers():
             imname = 'Xcorr_' + self.imageRaw_name
             if self.showXcorr.val and hasattr(self,'h'):
-                
                 im = self.get_current_stack_for_calibration()
-                ixf = np.squeeze(self.h.crossCorrelations(im))
+                # choose the gpu acceleration
+                if self.proc.currentData() == accel.USETORCH:
+                    ixf = np.squeeze(self.h.crossCorrelations_pytorch(im))
+                elif self.proc.currentData() == accel.USECUPY:
+                    ixf = np.squeeze(self.h.crossCorrelations_cupy(im))
+                else:
+                    ixf = np.squeeze(self.h.crossCorrelations(im))
+                # show the slected carrier
                 if ixf.ndim >2:
-                    phase_index = int(self.viewer.dims.current_step[1])
-                    carrier_idx = phase_index % ixf.shape[0]
-                    # if multiple carriers are found, the phase index is used to show it
-                    ixf=ixf[carrier_idx,:,:]
+                    carriers_number = ixf.shape[0]
+                    self.carrier_idx.set_min_max(0,carriers_number-1)
+                    carrier_idx = self.carrier_idx.val
+                    ixf =ixf [carrier_idx,:,:]   
+                else:
+                    self.carrier_idx.set_min_max(0,0)
                 self.show_image(ixf, imname, hold = True,
                                 colormap ='twilight', autoscale = True)
                 self.show_carrier()
@@ -550,9 +592,9 @@ class SimAnalysis(QWidget):
         '''
         dx = self.h.pixelsize / self.h.magnification  # Sampling in image plane
         res = self.h.wavelength / (2 * self.h.NA)
-        cutoff = 1/res/2 # coherent cutoff frequency
+        cutoff = 1 # coherent cutoff frequency
         oversampling = res / dx
-        dk = oversampling / (N / 2)  
+        dk = oversampling / (N / 2) # k space is normalised to coherent cutoff
         cutoff_in_pixels = cutoff / dk
         return cutoff_in_pixels, dk   
       
@@ -590,10 +632,10 @@ class SimAnalysis(QWidget):
             if self.showEta.val:
                 N = self.h.N
                 cutoff, dk   = self.calculate_kr(N)  
-                eta_radius = self.h.eta * cutoff
+                eta_radius = 1.9 * self.h.eta * cutoff
                 self.add_circles(np.array([N/2,N/2]), eta_radius,
                                name, color='green')
-                self.add_circles(np.array([N/2,N/2]), cutoff,
+                self.add_circles(np.array([N/2,N/2]), 2 * cutoff,
                                name, color='blue', hold=True)
             elif name in self.viewer.layers:
                 self.remove_layer(self.viewer.layers[name])
@@ -601,7 +643,6 @@ class SimAnalysis(QWidget):
     def add_circles(self, locations, radius=20,
                     shape_name='shapename', color='blue', hold=False
                     ):
-        
         '''
         Creates a circle in a layer with yx coordinates speciefied in each row of locations
         
@@ -663,13 +704,16 @@ class SimAnalysis(QWidget):
         sa,sp,sz,sy,sx = fullstack.shape
         phases_angles = sa*sp
         pa_stack = fullstack.reshape(phases_angles, sz, sy, sx)
-        demodulated = np.zeros([sz,sy,sx]).astype('float')
-        # if self.use_torch.val: TODO implement demodulation in torch, add to function call.astype(np.float32)
-        #     demodulation_function = self.h.OSreconstruct_pytorch
-        demodulation_function = self.h.filteredOSreconstruct
-        for frame_index in range(sz):
-            stack = np.squeeze(pa_stack[:,frame_index,:,:])
-            demodulated[frame_index,:,:] = np.squeeze(demodulation_function(stack))
+        paz_stack = np.swapaxes(pa_stack, 0, 1).reshape((phases_angles*sz, sy, sx))
+        if self.proc.currentData() == accel.USETORCH:
+            demodulation_function = self.h.filteredOSreconstruct_pytorch
+        elif self.proc.currentData() == accel.USECUPY:
+            demodulation_function = self.h.filteredOSreconstruct_cupy
+        else:
+            demodulation_function = self.h.filteredOSreconstruct
+        demodulated = demodulation_function(paz_stack)
+        if demodulated.ndim < 3:
+            demodulated = demodulated[np.newaxis, :]
         imname = 'Demodulated_' + self.imageRaw_name
         scale = [self.zscaling,1,1]
         self.show_image(demodulated, imname, scale=scale, hold=True, autoscale=True)
@@ -696,11 +740,16 @@ class SimAnalysis(QWidget):
         *args is to avoid conflic with the add_timer decorator
         '''
         if hasattr(self, 'h'):
-            imRaw = self.get_current_stack_for_calibration()         
-            if self.use_torch.val:
+            imRaw = self.get_current_stack_for_calibration()
+            start_time = time.time()
+            if self.proc.currentData() == accel.USETORCH:
                 self.h.calibrate_pytorch(imRaw,self.find_carrier.val)
+            elif self.proc.currentData() == accel.USECUPY:
+                self.h.calibrate_cupy(imRaw, self.find_carrier.val)
             else:
                 self.h.calibrate(imRaw,self.find_carrier.val)
+            elapsed_time = time.time() - start_time
+            self.messageBox.setText(f'Calibration time {elapsed_time:.3f}s')
             self.isCalibrated = True
             if self.find_carrier.val: # store the value found   
                 self.kx_input = self.h.kx  
@@ -722,8 +771,10 @@ class SimAnalysis(QWidget):
         dshape= current_image.shape
         phases_angles = self.phases_number.val*self.angles_number.val
         rdata = current_image.reshape(phases_angles, dshape[-2],dshape[-1])
-        if self.use_torch.val:
+        if self.proc.currentData() == accel.USETORCH:
             imageSIM = self.h.reconstruct_pytorch(rdata.astype(np.float32)) #TODO:this is left after conversion from torch
+        elif self.proc.currentData() == accel.USECUPY:
+            imageSIM = self.h.reconstruct_cupy(rdata.astype(np.float32))  # TODO:this is left after conversion from torch
         else:
             imageSIM = self.h.reconstruct_rfftw(rdata)
         imname = 'SIM_' + self.imageRaw_name
@@ -750,7 +801,7 @@ class SimAnalysis(QWidget):
                 phases_stack = np.squeeze(pa_stack[:,zidx,:,:])
                 if self.keep_calibrating.val:
                     delta = self.group.val // 2
-                    if zidx % delta == 0: 
+                    if (delta == 0) or (zidx % delta == 0):
                         remainer = self.group.val % 2
                         zmin = max(zidx-delta,0)
                         zmax = min(zidx+delta+remainer,sz)
@@ -758,12 +809,16 @@ class SimAnalysis(QWidget):
                         data = pa_stack[:,zmin:zmax,:,:]
                         s_pa = data.shape[0]
                         selected_imRaw = np.swapaxes(data, 0, 1).reshape((s_pa * new_delta, sy, sx))
-                        if self.use_torch.val:
+                        if self.proc.currentData() == accel.USETORCH:
                             self.h.calibrate_pytorch(selected_imRaw,self.find_carrier.val)
+                        elif self.proc.currentData() == accel.USECUPY:
+                            self.h.calibrate_cupy(selected_imRaw, self.find_carrier.val)
                         else:
                             self.h.calibrate(selected_imRaw,self.find_carrier.val)                
-                if self.use_torch.val:
+                if self.proc.currentData() == accel.USETORCH:
                     stackSIM[zidx,:,:] = self.h.reconstruct_pytorch(phases_stack.astype(np.float32)) #TODO:this is left after conversion from torch
+                elif self.proc.currentData() == accel.USECUPY:
+                    stackSIM[zidx, :, :] = self.h.reconstruct_cupy(phases_stack.astype(np.float32))  # TODO:this is left after conversion from torch
                 else:
                     stackSIM[zidx,:,:] = self.h.reconstruct_rfftw(phases_stack)      
             return stackSIM
@@ -771,13 +826,18 @@ class SimAnalysis(QWidget):
         @thread_worker(connect={'returned': update_sim_image})
         def _batch_reconstruction():
             warnings.filterwarnings('ignore')
-            if self.use_torch.val:
+            start_time = time.time()
+            if self.proc.currentData() == accel.USETORCH:
                 stackSIM = self.h.batchreconstructcompact_pytorch(paz_stack, blocksize = 32)
+            elif self.proc.currentData() == accel.USECUPY:
+                stackSIM = self.h.batchreconstructcompact_cupy(paz_stack, blocksize=32)
             else:
                 stackSIM = self.h.batchreconstructcompact(paz_stack)
+            elapsed_time = time.time() - start_time
+            self.messageBox.setText(f'Batch reconstruction time {elapsed_time:.3f}s')
             return stackSIM
         
-        # main function exetuted here
+        # main function executed here
         assert self.isCalibrated, 'SIM processor not calibrated, unable to perform SIM reconstruction'
         fullstack = self.get_hyperstack()
         sa,sp,sz,sy,sx = fullstack.shape
@@ -794,7 +854,7 @@ class SimAnalysis(QWidget):
         assert self.isCalibrated, 'SIM processor not calibrated, unable to show phases'
         if self.phases_number.val==7:
             self.find_hexsim_phaseshifts()
-        elif self.phases_number.val==3 :
+        elif self.phases_number.val==3 or self.phases_number.val==5 : #TODO make more general with combobox
             self.find_sim_phaseshifts()
         
         
@@ -806,9 +866,14 @@ class SimAnalysis(QWidget):
         sa,sp,sy,sx = stack.shape
         img = stack.reshape(sa*sp, sy, sx) 
         for i in range (3):
-            phase, _ = self.h.find_phase(self.h.kx[i], self.h.ky[i], img)
+            if self.proc.currentData() == accel.USETORCH:
+                phase, _ = self.h.find_phase_pytorch(self.h.kx[i], self.h.ky[i], img)
+            elif self.proc.currentData() == accel.USECUPY:
+                phase, _ = self.h.find_phase_cupy(self.h.kx[i], self.h.ky[i], img)
+            else:
+                phase, _ = self.h.find_phase(self.h.kx[i], self.h.ky[i], img)
             expected_phase[:,i] = np.arange(7) * 2*(i+1) * np.pi / 7
-            phaseshift[:,i] = np.unwrap(phase - phase[0])
+            phaseshift[:,i] = np.unwrap(phase - phase[0] - expected_phase[:,i]) + expected_phase[:,i]
         error = phaseshift-expected_phase
         data_to_plot = [expected_phase, phaseshift, error]
         symbols = ['.','o','|']
@@ -825,13 +890,18 @@ class SimAnalysis(QWidget):
             phaseshift = np.zeros((sp,sa))
             expected_phase = np.zeros((sp,sa))
             error = np.zeros((sp,sa))
-            phase, _ = self.h.find_phase(self.h.kx[angle_idx], self.h.ky[angle_idx], img)
+            if self.proc.currentData() == accel.USETORCH:
+                phase, _ = self.h.find_phase_pytorch(self.h.kx[angle_idx], self.h.ky[angle_idx], img)
+            elif self.proc.currentData() == accel.USECUPY:
+                phase, _ = self.h.find_phase_cupy(self.h.kx[angle_idx], self.h.ky[angle_idx], img)
+            else:
+                phase, _ = self.h.find_phase(self.h.kx[angle_idx], self.h.ky[angle_idx], img)
             phase = np.unwrap(phase)
             phase = phase.reshape(sa,sp).T
             expected_phase[:,angle_idx] = np.arange(sp) * 2*np.pi / sp
             phaseshift= phase-phase[0,:]
             error = phaseshift-expected_phase      
-            data_to_plot = [expected_phase, phaseshift, error]
+            data_to_plot = [expected_phase[:,angle_idx], phaseshift[:,angle_idx], error[:,angle_idx]]
             symbols = ['.','o','|']
             legend = ['expected', 'measured', 'error']
             self.plot_with_plt(data_to_plot, legend, symbols, title = f'angle {angle_idx}',
@@ -870,23 +940,19 @@ class SimAnalysis(QWidget):
         ticks = np.linspace(0, vmax*(vales_num-1)/vales_num, 2*vales_num-1 )
         ax.set_yticks(ticks)
         fig.tight_layout()
-        plt.show()
+        plt.show(block=False)
         plt.rcParams.update(plt.rcParamsDefault)
 
 
 if __name__ == '__main__':
-    
+
     import napari
-    
     viewer = napari.Viewer()
-    
     widget = SimAnalysis(viewer)
-    
     my_reshape_widget = reshape()    
     viewer.window.add_dock_widget(my_reshape_widget, name = 'Reshape stack', add_vertical_stretch = True)
     viewer.window.add_dock_widget(widget,
-                                  name = 'HexSim analyzer @Polimi',
+                                  name = 'Sim analyzer @Polimi',
                                   add_vertical_stretch = True)
-    
     napari.run() 
      
\ No newline at end of file
